#ifndef HELPERFUNCTIONS_H_
#define HELPERFUNCTIONS_H_

#include <termios.h>

#include <vector>
#include <algorithm>
#include <fstream>
#include <fcntl.h>
#include <linux/input.h>

#   define EMSG(fmt, args...) (fprintf( stderr, "%s(%d): " fmt "\n", __FILE__, __LINE__, ## args) && fflush(stderr))
#   define SMSG(fmt, args...) (fprintf( stdout, fmt "\n", ## args) && fflush(stdout))


extern EGLDisplay          m_eglDisplay;
extern EGLNativeWindowType m_eglWindow;
extern EGLContext          m_eglContext;
extern EGLSurface          m_eglSurface;


enum TextureFormat
{
    TEXTURE_ILLEGAL         	= -1,
    TEXTURE_A8R8G8B8			= 1,
    TEXTURE_R8G8B8				= 2,
    TEXTURE_R5G6B5				= 3,
    TEXTURE_A4R4G4B4			= 4,
    TEXTURE_GRAY				= 5,
    TEXTURE_GRAY_CLUT			= 6,
    TEXTURE_GRAY_ALPHA			= 7,
    TEXTURE_UYVY                = 8,
    TEXTURE_NV12                = 9,
    TEXTURE_YV12                = 10
};


// trim from start
inline std::string &ltrim(std::string &s) {
        s.erase(s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
        return s;
}

// trim from end
inline std::string &rtrim(std::string &s) {
        s.erase(std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
        return s;
}

// trim from both ends
inline std::string &trim(std::string &s) {
        return ltrim(rtrim(s));
}


inline int openKeyboard()
{
    struct Device
    {
        std::string id;
        std::string type;
        std::string name;
    };
    std::vector<Device> devices;

    std::ifstream devicesFile( "/proc/bus/input/devices", std::ifstream::in );

    Device parseDevice;
    std::string fullPath = "/dev/input/";
    std::string line;
    while( devicesFile.good() )
    {
        std::getline( devicesFile, line );
        trim(line);
        if( line.empty() )
        {
            if( !parseDevice.type.empty() )
                devices.push_back(parseDevice);
            parseDevice = Device();
        }
        else
        {
            if( line[0] == 'I' && line[1] == ':')
                parseDevice.id = line;
            else if( line[0] == 'B' && line[1] == ':' && line.find("EV=1200") != std::string::npos)
                parseDevice.type = line;
            else if(line[0] == 'H' && line[1] == ':')
            {
                size_t start = line.find("event");
                if( start != std::string::npos )
                {
                    size_t end = line.find( " ", start);
                    parseDevice.name = line.substr( start, end != std::string::npos ? end - start : end );
                }
            }
        }
    }
    if( !parseDevice.type.empty() )
        devices.push_back(parseDevice);

    int keyboardFD = open((fullPath + devices.front().name).c_str(), O_RDONLY);
    return keyboardFD;
}

inline bool eglErrorCheck()
{
    EGLint error = eglGetError();
    if( error == EGL_SUCCESS )
        return true;
    switch( error )
    {
    case EGL_BAD_ACCESS:
        EMSG(" Bad Access ");
        break;
    case EGL_BAD_ALLOC:
        EMSG(" Bad Allocation ");
        break;
    case EGL_BAD_ATTRIBUTE:
        EMSG(" Bad Attribute ");
        break;
    case EGL_NOT_INITIALIZED:
        EMSG(" Not initialized ");
        break;
    case EGL_BAD_CONFIG:
        EMSG(" EGL_BAD_CONFIG ");
        break;
    case EGL_BAD_CONTEXT:
        EMSG(" EGL_BAD_CONTEXT ");
        break;
    case EGL_BAD_CURRENT_SURFACE:
        EMSG(" EGL_BAD_CURRENT_SURFACE ");
        break;
    case EGL_BAD_DISPLAY:
        EMSG(" EGL_BAD_DISPLAY ");
        break;
    case EGL_BAD_MATCH:
        EMSG(" EGL_BAD_MATCH ");
        break;
    case EGL_BAD_NATIVE_PIXMAP:
        EMSG(" EGL_BAD_NATIVE_PIXMAP ");
        break;
    case EGL_BAD_NATIVE_WINDOW:
        EMSG(" EGL_BAD_NATIVE_WINDOW ");
        break;
    case EGL_BAD_PARAMETER:
        EMSG(" EGL_BAD_PARAMETER ");
        break;
    case EGL_BAD_SURFACE:
        EMSG(" EGL_BAD_SURFACE ");
        break;
    }
    return false;
}

#endif
