#include <EGL/egl.h>
#include <EGL/eglext.h>

#include <GL/gl.h>
#include <GL/glext.h>

#include <sys/types.h>
#include <scmem/lib/scmem.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>

#include "HelperFunctions.h"

PFNEGLCREATEIMAGEKHRPROC peglCreateImageKHR = 0;
PFNGLEGLIMAGETARGETTEXTURE2DOESPROC pglEGLImageTargetTexture2DOES = 0;
PFNEGLDESTROYIMAGEKHRPROC peglDestroyImageKHR = 0;
#ifdef EGL_MESA_platform_gbm
PFNEGLGETPLATFORMDISPLAYEXTPROC peglGetPlatformDisplayEXT = 0;
PFNEGLCREATEPLATFORMWINDOWSURFACEEXTPROC peglCreatePlatformWindowSurfaceEXT = 0;
PFNEGLCREATEPLATFORMPIXMAPSURFACEEXTPROC peglCreatePlatformPixmapSurfaceEXT = 0;
#endif

void initEGLFunctions()
{
    peglDestroyImageKHR = (PFNEGLDESTROYIMAGEKHRPROC)eglGetProcAddress("eglDestroyImageKHR");
    peglCreateImageKHR = (PFNEGLCREATEIMAGEKHRPROC)eglGetProcAddress("eglCreateImageKHR");
    pglEGLImageTargetTexture2DOES = (PFNGLEGLIMAGETARGETTEXTURE2DOESPROC)eglGetProcAddress("glEGLImageTargetTexture2DOES");

#ifdef EGL_MESA_platform_gbm
    peglGetPlatformDisplayEXT = (PFNEGLGETPLATFORMDISPLAYEXTPROC)eglGetProcAddress("eglGetPlatformDisplayExt");
    peglCreatePlatformWindowSurfaceEXT = (PFNEGLCREATEPLATFORMWINDOWSURFACEEXTPROC)eglGetProcAddress("eglCreatePlatformWindowSurfaceExt");
    peglCreatePlatformPixmapSurfaceEXT = (PFNEGLCREATEPLATFORMPIXMAPSURFACEEXTPROC)eglGetProcAddress("eglCreatePlatformPixmapSurfaceExt");
#endif
}

#ifndef X11

#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <gbm.h>
#include <drm.h>
#include <xf86drm.h>
#include <xf86drmMode.h>
#include <sys/poll.h>

static struct gbm_bo *g_bo = 0;
static int g_drm_fd = 0;
static struct gbm_device* g_gbm_device = 0;
static drmModeCrtcPtr g_orig_crtc = 0;
static drmModeRes *g_resources = 0;

EGLNativeDisplayType getNativeDisplay()
{
    if( g_gbm_device ) return (EGLNativeDisplayType) g_gbm_device;

    const char* card = getenv( "DRIDEV" );
    if( !card )
        card = "/dev/dri/card0";

    int fd = open(card, O_RDWR );
    if (fd < 0)
    {
        fprintf( stderr, "Error opening %s\n", card);
        return 0;
    }
    int ret = drmSetMaster(fd);
    if( ret != 0 )
    {
        fprintf( stderr, "Failed calling drmSetMaster: %s", strerror(errno));
        close(fd);
        return 0;
    }

    uint64_t has_dumb;
    if (drmGetCap(fd, DRM_CAP_DUMB_BUFFER, &has_dumb) < 0 || !has_dumb) {
        fprintf( stderr, "drm device '%s' does not support dumb buffers\n", card);
        close(fd);
        return 0;
    }

    g_gbm_device = gbm_create_device(fd);
    if (!g_gbm_device)
    {
        close(fd);
        fprintf( stderr, "Error opening GBM device\n" );
        return 0;
    }
    g_drm_fd = fd;

    return (EGLNativeDisplayType) g_gbm_device;
}

struct kms {
    drmModeConnector *connector;
    drmModeEncoder *encoder;
    drmModeModeInfoPtr mode;
    uint32_t crtc_id;
    uint32_t fb_id;
};

static struct kms g_kms;

struct drm_fb
{
    struct gbm_bo *bo;
    uint32_t fb_id;
};


static void drm_fb_destroy_callback(struct gbm_bo *bo, void *data)
{
    struct drm_fb *fb = (struct drm_fb *) data;
    struct gbm_device *gbm = gbm_bo_get_device(bo);

    if (fb->fb_id)
        drmModeRmFB(gbm_device_get_fd(gbm), fb->fb_id);
    free(fb);
}

static struct drm_fb *drm_fb_get_from_bo(struct gbm_bo *bo)
{
    struct drm_fb *fb = (struct drm_fb*)gbm_bo_get_user_data(bo);
    if (fb) return fb;
    fb = (struct drm_fb*)calloc(1, sizeof(*fb));
    fb->bo = bo;
    unsigned width = gbm_bo_get_width(bo);
    unsigned height = gbm_bo_get_height(bo);
    unsigned stride = gbm_bo_get_stride(bo);
    unsigned handle = gbm_bo_get_handle(bo).u32;
    fprintf( stdout, "[KMS/EGL]: New FB: %ux%u (stride: %u).\n", width, height, stride);
    int ret = drmModeAddFB(g_drm_fd, width, height, 24, 32, stride, handle, &fb->fb_id);
    if (ret < 0)
    {
        fprintf( stderr, "[KMS/EGL]: Failed to create FB: %s\n", strerror(errno));
        free(fb);
        return NULL;
    }

    gbm_bo_set_user_data(bo, fb, drm_fb_destroy_callback);
    return fb;
}

uint32_t current_fb_id = 0, next_fb_id = 0;

static void page_flip_handler(int /*fd*/, unsigned int /*frame*/, unsigned int /*sec*/, unsigned int /*usec*/, void *data)
{
    int *waiting_for_flip = (int*) data;
    *waiting_for_flip = 0;
}


void initDRM( EGLNativeWindowType window )
{
    if (!gbm_surface_has_free_buffers((struct gbm_surface *) window))
    {
        fprintf( stderr, "out of free buffers\n");
        return;
    }

    g_bo = gbm_surface_lock_front_buffer( (struct gbm_surface *) window);
    // fb will be destroyed in destroy callback of bo
    struct drm_fb *fb = drm_fb_get_from_bo(g_bo);
    /* set mode: */
    int ret = drmModeSetCrtc(g_drm_fd, g_kms.encoder->crtc_id, fb->fb_id, 0, 0, &g_kms.connector->connector_id, 1, g_kms.mode);
    if (ret) {
        printf("failed to set mode: %s\n", strerror(errno));
    }

}


int drmPageFlip(EGLNativeWindowType window)
{
    int ret;
    int waiting_for_flip = 1;
    struct gbm_bo *next_bo = gbm_surface_lock_front_buffer((struct gbm_surface *) window);
    // fb will be destroyed in destroy callback of bo
    struct drm_fb *fb = drm_fb_get_from_bo(next_bo);

    /*
    * Here you could also update drm plane layers if you want
    * hw composition
    */
    ret = drmModePageFlip(g_drm_fd, g_kms.encoder->crtc_id, fb->fb_id, DRM_MODE_PAGE_FLIP_EVENT, &waiting_for_flip);
    if (ret) {
        printf("failed to queue page flip: %s\n", strerror(errno));
        return -1;
    }

    drmEventContext evctx = {
        .version = DRM_EVENT_CONTEXT_VERSION,
        .vblank_handler = 0,
        .page_flip_handler = page_flip_handler
    };

    fd_set fds;
    while (waiting_for_flip)
    {
        FD_ZERO (&fds);
        FD_SET (g_drm_fd, &fds);
        ret = select(g_drm_fd+1, &fds, NULL, NULL, NULL);
        drmHandleEvent(g_drm_fd, &evctx);
    }
    /* release last buffer to render on again: */
    gbm_surface_release_buffer((struct gbm_surface *) window, g_bo);
    g_bo = next_bo;
    return 0;
}


static drmModePropertyPtr drm_get_prop(int fd, drmModeConnectorPtr connector, const char *name)
{
    for (int i = 0; i < connector->count_props; i++)
    {
        drmModePropertyPtr props = drmModeGetProperty(fd, connector->props[i]);
        if (!props || !props->name)
            continue;
        if (!strcmp(props->name, name))
            return props;
        drmModeFreeProperty(props);
    }
    return NULL;
}

static EGLBoolean setup_kms(int fd, struct kms *kms)
{    
    g_resources = drmModeGetResources(fd);
    if (!g_resources) {
        fprintf( stderr, "drmModeGetResources failed\n");
        return EGL_FALSE;
    }

    drmModeConnector *connector = 0;
    drmModeEncoder *encoder = 0;
    const char* displayTypeStr = getenv( "DISPLAY");
    uint32_t displayType = DRM_MODE_CONNECTOR_Unknown;
    if( displayTypeStr )
    {
        if( strcasecmp( displayTypeStr, "displayport") == 0 )
            displayType = DRM_MODE_CONNECTOR_DisplayPort;
        else if( strcasecmp( displayTypeStr, "vga") == 0 )
            displayType = DRM_MODE_CONNECTOR_VGA;
        else if( strcasecmp( displayTypeStr, "hdmia")== 0 )
            displayType = DRM_MODE_CONNECTOR_HDMIA;
        else if( strcasecmp( displayTypeStr, "hdmib") == 0 )
            displayType = DRM_MODE_CONNECTOR_HDMIB;
        else if( strcasecmp(displayTypeStr, "edp") == 0 )
            displayType = DRM_MODE_CONNECTOR_eDP;
        else if( strcasecmp(displayTypeStr, "dvid") == 0 )
            displayType = DRM_MODE_CONNECTOR_DVID;
    }
    int i;
    for (i = 0; i < g_resources->count_connectors; i++)
    {
        connector = drmModeGetConnector(fd, g_resources->connectors[i]);
        if (connector == NULL)
            continue;

        if( displayType && connector->connector_type != displayType )
        {
            fprintf( stdout,  "Skipping DisplayType: %d", connector->connector_type);
            continue;
        }

        if (connector->connection == DRM_MODE_CONNECTED && connector->count_modes > 0 )
            break;

        drmModeFreeConnector(connector);
    }

    drmModePropertyPtr prop = drm_get_prop(fd, connector, "DPMS");
    if (prop) {
        drmModeConnectorSetProperty(fd, connector->connector_id, prop->prop_id, DRM_MODE_DPMS_ON);
        drmModeFreeProperty(prop);
    }

    if (i == g_resources->count_connectors) {
        fprintf(stderr, "No currently active connector found.\n");
        return EGL_FALSE;
    }

    for (i = 0; i < g_resources->count_encoders; i++) {
        encoder = drmModeGetEncoder(fd, g_resources->encoders[i]);

        if (encoder == NULL)
            continue;

        if (encoder->encoder_id == connector->encoder_id)
            break;

        drmModeFreeEncoder(encoder);
    }

    kms->connector = connector;
    kms->encoder = encoder;
    kms->mode = &connector->modes[0];

    g_orig_crtc = drmModeGetCrtc(fd, kms->encoder->crtc_id);
    if (!g_orig_crtc)
        fprintf( stderr, "[KMS/EGL]: Cannot find original CRTC.\n");


    return EGL_TRUE;
}

EGLNativeWindowType getNativeWindow( EGLNativeDisplayType display )
{
    struct gbm_device *gbm = (struct gbm_device *) display;

    setup_kms( gbm_device_get_fd(gbm), &g_kms);

    struct gbm_surface *gs = gbm_surface_create(gbm, g_kms.mode->hdisplay, g_kms.mode->vdisplay, GBM_BO_FORMAT_XRGB8888, GBM_BO_USE_SCANOUT | GBM_BO_USE_RENDERING);

    //    drmModeFreeConnector(kms.connector);
    //    drmModeFreeEncoder(kms.encoder);

    return (EGLNativeWindowType) gs;
}

void releaseNativeWindow( EGLNativeWindowType& window, EGLNativeDisplayType display )
{
    if( g_drm_fd == 0 ) return;

    // Restore original CRTC.
    if (g_orig_crtc)
    {
        drmModeSetCrtc(g_drm_fd, g_orig_crtc->crtc_id,
                       g_orig_crtc->buffer_id,
                       g_orig_crtc->x,
                       g_orig_crtc->y,
                       &g_kms.connector->connector_id, 1, g_kms.mode);
        drmModeFreeCrtc(g_orig_crtc);
    }
    if( g_bo )
    {
        gbm_surface_release_buffer((struct gbm_surface *) window, g_bo);
        g_bo = 0;
    }
    gbm_surface_destroy((struct gbm_surface *)window);
    gbm_device_destroy( (struct gbm_device*)display);
    drmModeFreeEncoder(g_kms.encoder);
    drmModeFreeConnector(g_kms.connector);
    if (g_resources)
        drmModeFreeResources(g_resources);

    int ret = drmDropMaster(g_drm_fd);
    if( ret != 0 )
        fprintf( stderr, "Failed calling drmDropMaster: %s", strerror(errno));

    close( g_drm_fd );
    g_drm_fd = 0;

    g_gbm_device = 0;
}

#endif

bool initEGL( EGLNativeDisplayType nativeDisplay, EGLNativeWindowType window, TextureFormat& format, int depth, int samples)
{
    EGLConfig config;
    EGLDisplay display = eglGetDisplay( nativeDisplay );
    if ( display == EGL_NO_DISPLAY || !eglErrorCheck() )
    {
        fprintf( stderr, "Invalid Display\n" );
        return false;
    }

    m_eglDisplay = display;
    m_eglWindow = window;

    EGLint majorVersion;
    EGLint minorVersion;
    EGLint numConfigs;

    // Initialize EGL
    if ( !eglInitialize(m_eglDisplay, &majorVersion, &minorVersion) || !eglErrorCheck()  )
    {
        fprintf( stderr, "Error initializing egl)\n");
        return false;
    }

    if( !eglBindAPI(EGL_OPENGL_ES_API) || !eglErrorCheck() )
    {
        fprintf( stderr, "Error binding EGL_OPENGL_ES_API\n" );
        return false;
    }

    //printEGLConfigs( display );

    EGLint attribList[] = {
        EGL_SURFACE_TYPE,         EGL_WINDOW_BIT, // | EGL_PIXMAP_BIT,
        EGL_RED_SIZE,             8,
        EGL_GREEN_SIZE,           8,
        EGL_BLUE_SIZE,            8,
        EGL_ALPHA_SIZE,           8,
        EGL_LUMINANCE_SIZE,       0,
        EGL_DEPTH_SIZE,           depth,
        EGL_SAMPLE_BUFFERS,       samples > 0,
        EGL_SAMPLES,              samples,
        EGL_RENDERABLE_TYPE,      EGL_OPENGL_ES2_BIT,
        EGL_NONE,                 EGL_NONE
    };


    switch( format )
    {
    case TEXTURE_R5G6B5:
        // EGL_RED_SIZE
        attribList[3] = 5;
        // EGL_GREEN_SIZE
        attribList[5] = 6;
        // EGL_BLUE_SIZE
        attribList[7] = 5;
        // EGL_ALPHA_SIZE
        attribList[9] = 0;
        break;
    case TEXTURE_A8R8G8B8:
        // EGL_RED_SIZE
        attribList[3] = 8;
        // EGL_GREEN_SIZE
        attribList[5] = 8;
        // EGL_BLUE_SIZE
        attribList[7] = 8;
        // EGL_ALPHA_SIZE
        attribList[9] = 8;
        break;
    case TEXTURE_GRAY:
        fprintf( stderr, "Error: Image m_externalFormat is not supported yet\n" );
        // TODO use EGL_COLOR_BUFFER_TYPE EGL_LUMINANCE_BUFFER
        break;
    case TEXTURE_GRAY_ALPHA:
        fprintf( stderr, "Error: Image m_externalFormat is not supported yet\n" );
        break;
    case TEXTURE_A4R4G4B4:
        fprintf( stderr, "Error: Image m_externalFormat is not supported yet\n" );
        break;
    case TEXTURE_R8G8B8:
        fprintf( stderr, "Error: Image m_externalFormat is not supported yet\n" );
        break;
    default:
        fprintf( stderr, "Error: Image m_externalFormat is not supported yet\n" );
        return false;
    }


    // Choose config
    if ( !eglChooseConfig(m_eglDisplay, attribList, &config, 1, &numConfigs) || !eglErrorCheck() || numConfigs != 1 )
    {
        fprintf( stderr, "Error choosing config for specified EGL attribute list\n");
        return false;
    }


    printf("Configurations:\n");
    printf("     bf lv d st colorbuffer dp st  ms    vis   supported renderable\n");
    printf("  id sz  l b ro  r  g  b  a th cl ns b    id   surfaces        type\n");
    printf("-------------------------------------------------------------------\n");

    EGLint id = 0, size = 0, level = 0;
    EGLint red = 0, green = 0, blue = 0, alpha = 0;
    EGLint depthBpp = 0, stencil = 0;
    EGLint surfaces = 0;
    EGLint doubleBuf = 1, stereo = 0;
    EGLint vid = 0;
    EGLint gotSamples = 0, sampleBuffers = 0;
    EGLint renderableType = 0;
    EGLint contextClientVersion = 0;

    char surfString[100] = "";

    eglGetConfigAttrib( m_eglDisplay, config, EGL_CONFIG_ID, &id);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_BUFFER_SIZE, &size);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_LEVEL, &level);

    eglGetConfigAttrib( m_eglDisplay, config, EGL_RED_SIZE, &red);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_GREEN_SIZE, &green);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_BLUE_SIZE, &blue);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_ALPHA_SIZE, &alpha);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_DEPTH_SIZE, &depthBpp);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_STENCIL_SIZE, &stencil);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_NATIVE_VISUAL_ID, &vid);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_SURFACE_TYPE, &surfaces);

    eglGetConfigAttrib( m_eglDisplay, config, EGL_SAMPLES, &gotSamples);
    eglGetConfigAttrib( m_eglDisplay, config, EGL_SAMPLE_BUFFERS, &sampleBuffers);

    if (!eglGetConfigAttrib(display, config, EGL_RENDERABLE_TYPE, &renderableType)) {
        fprintf( stderr, "can't query EGLConfig RENDERABLE_TYPE\n");
    }

    if (renderableType & EGL_OPENGL_ES2_BIT) {
        contextClientVersion = 2;
    } else if (renderableType & EGL_OPENGL_ES_BIT) {
        contextClientVersion = 1;
    } else {
        fprintf( stderr, "no supported EGL_RENDERABLE_TYPEs\n");
    }

    if (surfaces & EGL_WINDOW_BIT)
        strcat(surfString, "win,");
    if (surfaces & EGL_PBUFFER_BIT)
        strcat(surfString, "pb,");
    if (surfaces & EGL_PIXMAP_BIT)
        strcat(surfString, "pix,");
    if (strlen(surfString) > 0)
        surfString[strlen(surfString) - 1] = 0;

    printf("0x%02x %2d %2d %c  %c %2d %2d %2d %2d %2d %2d %2d%2d  0x%02x   %-12s      %2d\n",
           id, size, level,
           doubleBuf ? 'y' : '.',
           stereo ? 'y' : '.',
           red, green, blue, alpha,
           depthBpp, stencil,
           gotSamples, sampleBuffers, vid, surfString, contextClientVersion);


    // Create a GL context
    EGLint contextAttribs[] = { EGL_CONTEXT_CLIENT_VERSION, 2, EGL_NONE, EGL_NONE };
    if( red == 8 && blue == 8 && green == 8 && alpha == 8 )
        format = TEXTURE_A8R8G8B8;
    else if( red == 5 && blue == 5 && green == 6 && alpha == 0 )
        format = TEXTURE_R5G6B5;

    m_eglContext = eglCreateContext(m_eglDisplay, config, EGL_NO_CONTEXT, contextAttribs );
    if ( !eglErrorCheck() || m_eglContext == EGL_NO_CONTEXT )
    {
        fprintf( stderr, "Error creating EGL context\n");
        return false;
    }

    // Create a surface
    EGLint		attrib_list[16] = {EGL_NONE};
    m_eglSurface = eglCreateWindowSurface(m_eglDisplay, config, m_eglWindow, attrib_list);
    if ( m_eglSurface == EGL_NO_SURFACE || !eglErrorCheck()  )
    {
        fprintf( stderr, "Error creating window surface\n");
        return false;
    }

    if( !eglQueryContext( m_eglDisplay, m_eglContext, EGL_CONTEXT_CLIENT_VERSION, &contextClientVersion ) )
        fprintf( stderr, "can't query EGL_CONTEXT_CLIENT_VERSION\n");
    else
        fprintf( stdout, "Context created: Version %d\n", contextClientVersion );
    return true;
}


void create_native_pixmap( TextureFormat pixmapFormat, unsigned long pixmapwidth, unsigned long pixmapHeight, unsigned long long* physicalAddress )
{
    if( !physicalAddress ) return;

    fprintf (stderr, "create_native_pixmap: entered\n");
    *physicalAddress = 0;
    int bytesPerPixel = 1;
    //The scmem module should be inserted before running this function
    switch(pixmapFormat)
    {
    case TEXTURE_R5G6B5:
    case TEXTURE_A4R4G4B4:
    case TEXTURE_UYVY:
        bytesPerPixel = 2;
        break;
    case TEXTURE_R8G8B8:
        bytesPerPixel = 3;
        break;
    case TEXTURE_A8R8G8B8:
        bytesPerPixel = 4;
        break;
    default:
        fprintf( stderr, "Invalid pixmap format type %d\n", pixmapFormat);
        return;
    }

    if( scmem_init() != 0 )
        return;

    fprintf (stderr, "create_native_pixmap: scmem initialized\n");

    int sizeInBytes = pixmapwidth * pixmapHeight * bytesPerPixel;

    // Try page align to avoid problems
    const int pageSize = getpagesize();
    while( sizeInBytes % pageSize )
        ++sizeInBytes;

    fprintf (stderr, "create_native_pixmap: Looking to allocate %#08x bytes from CMEM pool\n", sizeInBytes);

    *physicalAddress = scmem_alloc(sizeInBytes, 1);
    if(!*physicalAddress)
    {
        fprintf( stderr, "scmem_alloc returned NULL\n");
        return;
    }    
}


void release_native_pixmap( unsigned long long physicalAddress)
{
    scmem_free( physicalAddress );
    scmem_deinit();
}
