#include <OpenGL/OpenGL.h>

#include <Basic/Logging.h>
#include <string.h>

bool eglErrorCheck()
{
    EGLint error = eglGetError();
    if( error == EGL_SUCCESS )
        return true;
    switch( error )
    {
        case EGL_BAD_ACCESS:
            EMSG(" Bad Access ");
            break;
        case EGL_BAD_ALLOC:
            EMSG(" Bad Allocation ");
            break;
        case EGL_BAD_ATTRIBUTE:
            EMSG(" Bad Attribute ");
            break;
        case EGL_NOT_INITIALIZED:
            EMSG(" Not initialized ");
            break;
        case EGL_BAD_CONFIG:
            EMSG(" EGL_BAD_CONFIG ");
            break;
        case EGL_BAD_CONTEXT:
            EMSG(" EGL_BAD_CONTEXT ");
            break;
        case EGL_BAD_CURRENT_SURFACE:
            EMSG(" EGL_BAD_CURRENT_SURFACE ");
            break;
        case EGL_BAD_DISPLAY:
            EMSG(" EGL_BAD_DISPLAY ");
            break;
        case EGL_BAD_MATCH:
            EMSG(" EGL_BAD_MATCH ");
            break;
        case EGL_BAD_NATIVE_PIXMAP:
            EMSG(" EGL_BAD_NATIVE_PIXMAP ");
            break;
        case EGL_BAD_NATIVE_WINDOW:
            EMSG(" EGL_BAD_NATIVE_WINDOW ");
            break;
        case EGL_BAD_PARAMETER:
            EMSG(" EGL_BAD_PARAMETER ");
            break;
        case EGL_BAD_SURFACE:
            EMSG(" EGL_BAD_SURFACE ");
            break;
    }
    return false;
}

#define MAX_CONFIGS 1000
/**
 * Print table of all available configurations.
 */
void printEGLConfigs(EGLDisplay d)
{
   EGLConfig configs[MAX_CONFIGS];
   EGLint numConfigs, i;

   eglGetConfigs(d, configs, MAX_CONFIGS, &numConfigs);

   printf("Configurations:\n");
   printf("     bf lv d st colorbuffer dp st  ms    vis   supported\n");
   printf("  id sz  l b ro  r  g  b  a th cl ns b    id   surfaces \n");
   printf("--------------------------------------------------------\n");
   for (i = 0; i < numConfigs; i++) {
      EGLint id, size, level;
      EGLint red, green, blue, alpha;
      EGLint depth, stencil;
      EGLint surfaces;
      EGLint doubleBuf = 1, stereo = 0;
      EGLint vid;
      EGLint samples, sampleBuffers;
      char surfString[100] = "";

      eglGetConfigAttrib(d, configs[i], EGL_CONFIG_ID, &id);
      eglGetConfigAttrib(d, configs[i], EGL_BUFFER_SIZE, &size);
      eglGetConfigAttrib(d, configs[i], EGL_LEVEL, &level);

      eglGetConfigAttrib(d, configs[i], EGL_RED_SIZE, &red);
      eglGetConfigAttrib(d, configs[i], EGL_GREEN_SIZE, &green);
      eglGetConfigAttrib(d, configs[i], EGL_BLUE_SIZE, &blue);
      eglGetConfigAttrib(d, configs[i], EGL_ALPHA_SIZE, &alpha);
      eglGetConfigAttrib(d, configs[i], EGL_DEPTH_SIZE, &depth);
      eglGetConfigAttrib(d, configs[i], EGL_STENCIL_SIZE, &stencil);
      eglGetConfigAttrib(d, configs[i], EGL_NATIVE_VISUAL_ID, &vid);
      eglGetConfigAttrib(d, configs[i], EGL_SURFACE_TYPE, &surfaces);

      eglGetConfigAttrib(d, configs[i], EGL_SAMPLES, &samples);
      eglGetConfigAttrib(d, configs[i], EGL_SAMPLE_BUFFERS, &sampleBuffers);

      if (surfaces & EGL_WINDOW_BIT)
         strcat(surfString, "win,");
      if (surfaces & EGL_PBUFFER_BIT)
         strcat(surfString, "pb,");
      if (surfaces & EGL_PIXMAP_BIT)
         strcat(surfString, "pix,");
#ifdef EGL_MESA_screen_surface
      if (surfaces & EGL_SCREEN_BIT_MESA)
         strcat(surfString, "scrn,");
#endif
      if (strlen(surfString) > 0)
         surfString[strlen(surfString) - 1] = 0;

      printf("0x%02x %2d %2d %c  %c %2d %2d %2d %2d %2d %2d %2d%2d  0x%02x   %-12s\n",
             id, size, level,
             doubleBuf ? 'y' : '.',
             stereo ? 'y' : '.',
             red, green, blue, alpha,
             depth, stencil,
             samples, sampleBuffers, vid, surfString);
   }
}

bool isExtensionSupported(EGLDisplay dpy, const char *extension)
{
    const char *extensions = NULL;
    const char *start;
    char *where, *terminator;

    /* Extension names should not have spaces. */
    where = (char *) strchr(extension, ' ');
    if (where || *extension == '\0')
        return false;

    extensions = eglQueryString(dpy, EGL_EXTENSIONS);

    /* It takes a bit of care to be fool-proof about parsing the
    OpenGL extensions string. Don't be fooled by sub-strings, etc. */
    start = extensions;
    for (;;)
    {
        where = (char *) strstr((const char *) start, extension);
        if (!where)
            break;
        terminator = where + strlen(extension);
        if (where == start || *(where - 1) == ' ')
            if (*terminator == ' ' || *terminator == '\0')
                return true;
        start = terminator;
    }
    return false;
}

#ifndef NDEBUG
void glErrorCheck( const char* file /*=0*/, const int line /*= 0*/)
{
    GLenum error = glGetError();
    if(error != GL_NO_ERROR)
    {
        switch( error )
        {
        case GL_INVALID_OPERATION:
            SMSG ("%s( %d ): OpenGL: Invalid Operation Error! %d\n", file, line, error );
            break;
        case GL_INVALID_ENUM:
            SMSG ("%s( %d ): OpenGL: Invalid Enum Error! %d\n", file, line, error );
            break;
        case GL_INVALID_VALUE:
            SMSG ("%s( %d ): OpenGL: Invalid Value Error! %d\n", file, line, error );
            break;
        case GL_OUT_OF_MEMORY:
            SMSG ("%s( %d ): OpenGL: Out of memory error! %d\n", file, line, error );
            break;
        }        
    }    
}
#endif

void showConfigs( EGLDisplay dsp, EGLConfig* configList, EGLint configSize )
{
    EGLint value, r, g, b, a;
    for( int i = 0; i < configSize; ++i )
    {
        SMSG("\nEGL Configuration %d:", i);
        eglGetConfigAttrib(dsp, configList[i], EGL_BUFFER_SIZE, &value);
        SMSG("Overall Buffer Depth = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_RED_SIZE, &r);
        eglGetConfigAttrib(dsp, configList[i], EGL_GREEN_SIZE, &g);
        eglGetConfigAttrib(dsp, configList[i], EGL_BLUE_SIZE, &b);
        eglGetConfigAttrib(dsp, configList[i], EGL_ALPHA_SIZE, &a);
        SMSG("Color Buffer Depth: Red = %d Green %d Blue %d Alpha %d", r, g, b, a);
        eglGetConfigAttrib(dsp, configList[i], EGL_CONFIG_CAVEAT, &value);
        switch( value )
        {
        case EGL_NONE:
            SMSG("No Caveat in this Configuration");
            break;
        case EGL_SLOW_CONFIG:
            SMSG("Slow Config Caveat in this Configuration");
            break;
        case EGL_NON_CONFORMANT_CONFIG:
            SMSG("Non Conformant Caveat in this Configuration");
            break;
        }
        eglGetConfigAttrib(dsp, configList[i], EGL_CONFIG_ID, &value);
        SMSG("Config ID = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_DEPTH_SIZE, &value);
        SMSG("Depth Buffer Size = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_LEVEL, &value);
        SMSG("Frame Buffer Level = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_MAX_PBUFFER_HEIGHT, &r);
        eglGetConfigAttrib(dsp, configList[i], EGL_MAX_PBUFFER_WIDTH, &g);
        eglGetConfigAttrib(dsp, configList[i], EGL_MAX_PBUFFER_PIXELS, &b);
        SMSG("Max Pbuffer: Height = %d Width %d Pixels %d", r, g, b);
        eglGetConfigAttrib(dsp, configList[i], EGL_NATIVE_RENDERABLE, &value);
        SMSG("Native Renderable = %s", value == EGL_TRUE ? "true" : "false" );
        eglGetConfigAttrib(dsp, configList[i], EGL_NATIVE_VISUAL_ID, &value);
        SMSG("Native Visual ID = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_NATIVE_VISUAL_TYPE, &value);
        SMSG("Native Visual Type = %d", value);
        //eglGetConfigAttrib(dsp, configList[i], EGL_PRESERVED_RESOURCES, &value);
        //SMSG("Resources are preserved accross power management = %s", value == EGL_TRUE ? L"true" : L"false");
        eglGetConfigAttrib(dsp, configList[i], EGL_SAMPLE_BUFFERS, &value);
        SMSG("Multisample Buffers = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_SAMPLES, &value);
        SMSG("Samples per pixel = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_STENCIL_SIZE, &value);
        SMSG("Number of bits in stencil buffer = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_SURFACE_TYPE, &value);
        if( value & EGL_WINDOW_BIT)
            SMSG("Surface Type = EGL_WINDOW_BIT");
        if( value & EGL_PBUFFER_BIT)
            SMSG("Surface Type = EGL_PBUFFER_BIT");
        if( value & EGL_PIXMAP_BIT )
            SMSG("Surface Type = EGL_PIXMAP_BIT");
        eglGetConfigAttrib(dsp, configList[i], EGL_TRANSPARENT_TYPE, &value);
        SMSG("Transparent Type = %s", value == EGL_NONE ? "None" : "RGB");
        eglGetConfigAttrib(dsp, configList[i], EGL_TRANSPARENT_RED_VALUE, &value);
        SMSG("Red Transparent = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_TRANSPARENT_GREEN_VALUE, &value);
        SMSG("Green Transparent = %d", value);
        eglGetConfigAttrib(dsp, configList[i], EGL_TRANSPARENT_BLUE_VALUE, &value);
        SMSG("Blue Transparent = %d", value);
    }
}
