#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/moduleparam.h>
#include <linux/list.h>
#include <linux/cdev.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/seq_file.h>
#include <linux/scatterlist.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <asm/cacheflush.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>

#include <linux/version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 12, 0)
#include <linux/dma-buf.h>
#endif

#ifdef CONFIG_DMA_SHARED_BUFFER
#include <linux/dma-contiguous.h>
#endif


#ifndef VM_RESERVED
#define VM_RESERVED 0x00080000
#endif

#include "scmemk.h"

/**
  * This module uses dev_dbg for debug output, see https://lwn.net/Articles/434833/ for details how to specify what output should be printed
  */

/* Register the module parameters. */
MODULE_PARM_DESC(phys_start, "\n\t\t Start Address for SCMEM Pool Memory");
static char *phys_start = NULL;
MODULE_PARM_DESC(phys_end, "\n\t\t End Address for SCMEM Pool Memory");
static char *phys_end = NULL;
module_param(phys_start, charp, S_IRUGO);
module_param(phys_end, charp, S_IRUGO);

static int allowOverlap = 0;
MODULE_PARM_DESC(allowOverlap,
                 "\n\t\t Set to 1 if scmem range is allowed to overlap memory range"
                 "\n\t\t allocated to kernel physical mem (via mem=xxx)");
module_param(allowOverlap, int, S_IRUGO);


/* Forward declaration of system calls */
static long ioctl(struct file *filp, unsigned int cmd, unsigned long args);
static int mmap(struct file *filp, struct vm_area_struct *vma);
static int open(struct inode *inode, struct file *filp);
static int release(struct inode *inode, struct file *filp);

static struct file_operations scmem_fxns = {
    owner:   THIS_MODULE,
    unlocked_ioctl: ioctl,
                    mmap:    mmap,
                    open:    open,
                    release: release
};


#ifdef CONFIG_DMA_SHARED_BUFFER
    // DMA_BUF API
    static int dmabuf_mmap(struct dma_buf *dmabuf, struct vm_area_struct *vma);
//    static int dmabuf_attach(struct dma_buf *, struct device *, struct dma_buf_attachment *);
//    static void dmabuf_detach(struct dma_buf *, struct dma_buf_attachment *);
    static struct sg_table *dmabuf_map_dma_buf(struct dma_buf_attachment *attachment, enum dma_data_direction direction);
    static void dmabuf_unmap_dma_buf(struct dma_buf_attachment *attachment, struct sg_table *table, enum dma_data_direction direction);
    static void dmabuf_dma_buf_release(struct dma_buf *dmabuf);
    static void *dmabuf_dma_buf_kmap(struct dma_buf *dmabuf, unsigned long offset);
    static void dmabuf_dma_buf_kunmap(struct dma_buf *dmabuf, unsigned long offset, void *ptr);
    static int dmabuf_dma_buf_begin_cpu_access(struct dma_buf *dmabuf, size_t start, size_t len, enum dma_data_direction direction);
    static void dmabuf_dma_buf_end_cpu_access(struct dma_buf *dmabuf, size_t start, size_t len, enum dma_data_direction direction);

    static struct dma_buf_ops dma_buf_ops = {
//        .owner = THIS_MODULE,
//        .attach = dmabuf_attach,
//        .detach = dmabuf_detach,
        .map_dma_buf = dmabuf_map_dma_buf,
        .unmap_dma_buf = dmabuf_unmap_dma_buf,
        .mmap = dmabuf_mmap,
        .release = dmabuf_dma_buf_release,
        .begin_cpu_access = dmabuf_dma_buf_begin_cpu_access,
        .end_cpu_access = dmabuf_dma_buf_end_cpu_access,
        .kmap_atomic = dmabuf_dma_buf_kmap,
        .kunmap_atomic = dmabuf_dma_buf_kunmap,
        .kmap = dmabuf_dma_buf_kmap,
        .kunmap = dmabuf_dma_buf_kunmap,
    };

#endif


#define BLOCK_MEMREGION  (1 << 1)

    /* Describes a pool buffer */
    typedef struct pool_buffer {
        struct list_head element;
        struct list_head dma_users;
        struct list_head uspace_users;
        int id;
        phys_addr_t physp;
        int flags;			/* SCMEM_CACHED or SCMEM_NONCACHED */
        unsigned long long size;	/* used only for heap-based allocs */
        struct dma_buf *dmabuf;
        struct sg_table table;
    } pool_buffer;

    /* Describes a pool */
    typedef struct pool_object {
        struct list_head freelist;
        struct list_head busylist;
        unsigned int numbufs;
        //unsigned long long size;
        unsigned long long reqsize;
    } pool_object;

    typedef struct registered_user {
        struct list_head element;
        void *user;
    } registered_user;


    struct device*                  g_scmem_dev = 0;
    static struct mutex             g_scmem_mutex;
    static struct pool_object       g_allocations;
    static int                      g_scmem_major;
    static struct class*            g_scmem_class = 0;

    static unsigned long long       g_aligned_physstart = 0;
    static unsigned long long       g_aligned_physend = 0;
    static unsigned long long       g_available_mem = 0;
    static unsigned int             g_mem_flags = 0;

    static atomic_t                 g_reference_count = ATOMIC_INIT(0);


    /* Traverses the page tables and translates a virtual address to a physical. */
    static phys_addr_t get_phys(void *virtp)
    {
        unsigned long virt = (unsigned long)virtp;
        phys_addr_t physp = ~(0LL);
        struct mm_struct *mm = current->mm;
        struct vm_area_struct *vma;

        /* For kernel direct-mapped memory, take the easy way */
        if (virt >= PAGE_OFFSET)
        {
            physp = virt_to_phys(virtp);
            dev_dbg(g_scmem_dev, "get_phys: virt_to_phys translated direct-mapped %#lx to %#llx\n", virt, (unsigned long long)physp);
        }

        /* this will catch, kernel-allocated, mmaped-to-usermode addresses */
        else if ((vma = find_vma(mm, virt)) && (vma->vm_flags & VM_IO) && (vma->vm_pgoff))
        {
            dev_dbg(g_scmem_dev, "get_phys: sizeof(virt)  = %zu\n", sizeof(virt));
            dev_dbg(g_scmem_dev, "get_phys: virtp = %#lx\n", virt);
            dev_dbg(g_scmem_dev, "get_phys: vma->vm_start     = %#lx\n", vma->vm_start);
            dev_dbg(g_scmem_dev, "get_phys: vma->vm_end       = %#lx\n", vma->vm_end);
            dev_dbg(g_scmem_dev, "get_phys: vma->vm_pgoff     = %#lx\n", vma->vm_pgoff);
            physp = ((unsigned long long)vma->vm_pgoff << PAGE_SHIFT) + (virt - vma->vm_start);
            dev_dbg(g_scmem_dev, "get_phys: find_vma translated user %#lx to %#llx\n", virt, (unsigned long long)physp);
        }

        /* otherwise, use get_user_pages() for general userland pages */
        else
        {
            int res, nr_pages = 1;
            struct page *pages;

            down_read(&current->mm->mmap_sem);
            res = get_user_pages(current, current->mm, virt, nr_pages, 1, 0,
                                 &pages, NULL);
            up_read(&current->mm->mmap_sem);

            if (res == nr_pages) {
                physp = __pa(page_address(&pages[0]) + (virt & ~PAGE_MASK));
                dev_dbg(g_scmem_dev, "get_phys: get_user_pages translated user %#lx to %#llx\n", virt, (unsigned long long)physp);
            } else {
                printk( KERN_ERR "%s: Unable to find phys addr for %#lx\n", __FUNCTION__, virt);
                printk( KERN_ERR "%s: get_user_pages() failed: %d\n", __FUNCTION__, res);
            }
        }

        return physp;
    }

    /*
 *  ======== find_busy_entry ========
 *  find_busy_entry looks for an allocated pool buffer containing
 *  physical addr physp -> (physp + *sizep).
 *
 *  Should be called with the scmem_mutex held.
 */
    static struct pool_buffer *find_busy_entry(phys_addr_t physp, int *poolp, struct list_head **ep, unsigned long long* sizep)
    {
        struct list_head *busylistp;
        struct list_head *e;
        struct pool_buffer *entry;

        // size with zero is equal to no sizep given
        if( sizep && !*sizep )
            sizep = 0;

        busylistp = &g_allocations.busylist;
        for (e = busylistp->next; e != busylistp; e = e->next)
        {
            entry = list_entry(e, struct pool_buffer, element);
            if( (!sizep && entry->physp == physp) || (sizep && physp == entry->physp && (*sizep == entry->size) ) )
            {
                if (poolp) {
                    *poolp = entry->id;
                }
                if (ep) {
                    *ep = e;
                }
                printk (KERN_INFO "%s: physp=%#llx size=%llx entry->physp=%#llx entry->size=%#llx.\n", __func__, (unsigned long long)physp, sizep ? *sizep : 0, (unsigned long long) entry->physp, entry->size);
                return entry;
            }
        }
        printk (KERN_INFO "%s: physp=%#llx size=%llx not found\n", __func__, (unsigned long long)physp, sizep ? *sizep : 0);
        return NULL;
    }

    static int is_last_buffer( struct pool_buffer* entry )
    {
        struct list_head *busylistp;
        struct list_head *e;
        struct pool_buffer* elem;

        busylistp = &g_allocations.busylist;
        for (e = busylistp->next; e != busylistp; e = e->next)
        {
            elem = list_entry(e, struct pool_buffer, element);
            if( elem->physp > entry->physp )
                return 0;
        }
        dev_dbg( g_scmem_dev, "entry %d (%#llx) seems to be the last buffer\n", entry->id, (unsigned long long) entry->physp );
        return 1;
    }

    static void release_free_buffers(void)
    {
        struct list_head *freelistp = &g_allocations.freelist;
        struct list_head *busylistp = &g_allocations.busylist;
        struct list_head *e, *q;
        struct pool_buffer* entry;

        // Make sure that there is no buffer left in use with a physp greater than physpStart

        phys_addr_t lastUsed = 0;
        for (e = busylistp->next; e != busylistp; e = e->next)
        {
            entry = list_entry(e, struct pool_buffer, element);
            if( lastUsed < entry->physp )
                lastUsed = entry->physp;
        }

        e = freelistp->next;
        // First search any free memory block allocated before
        list_for_each_safe(e, q, freelistp )
        {
            entry = list_entry(e, struct pool_buffer, element);            
            if( entry->physp >= lastUsed )
            {
                g_available_mem += entry->size;
                g_allocations.reqsize -= entry->size;
                dev_dbg( g_scmem_dev, "Removing entry from freelist %#llx\n", (unsigned long long) entry->physp);
                list_del(e);
                // Make sure we don't reference the freed buffer anymore if a pending dma_buf mapping is alive in some way
#ifdef CONFIG_DMA_SHARED_BUFFER
                if( entry->dmabuf )
                    entry->dmabuf->priv = 0;
#endif
                kfree(entry);
            }
        }
    }

    /* Allocates space from the contiguous buffer for pool buffer. */
    static struct pool_buffer * allocBuffer(unsigned long long size)
    {
        struct pool_buffer *entry = 0;
        struct list_head *freelistp = &g_allocations.freelist;
        struct list_head *busylistp = &g_allocations.busylist;
        struct list_head *e = freelistp->next;

        dev_dbg( g_scmem_dev, "alloc_pool_buffer: Called for size %llu\n", size);

        // First search any free memory block allocated before
        while (e != freelistp) {
            entry = list_entry(e, struct pool_buffer, element);
            dev_dbg( g_scmem_dev, "Allocated buffer %d, physical %#llx and size %#llx from freelist\n", entry->id, (unsigned long long)entry->physp, size);
            // Fits, take it
            if( entry->size > size )
            {
                struct pool_buffer *newEntry = kzalloc(sizeof(struct pool_buffer), GFP_KERNEL);
                if (!newEntry)
                    printk( KERN_ERR "alloc_pool_buffer failed to malloc pool_buffer struct for remaining memory");
                else
                {
                    newEntry->size = entry->size - size;
                    newEntry->physp = entry->physp + size;
                    newEntry->id = g_allocations.numbufs++;
                    INIT_LIST_HEAD( &newEntry->dma_users );
                    INIT_LIST_HEAD( &newEntry->uspace_users );
                    list_add_tail(&newEntry->element, freelistp);
                    entry->size = size;
                }
                list_del_init(e);
                list_add(e, busylistp);
                return entry;
            }
            e = e->next;
        }
        // Check remaining memory
        if (size <= g_available_mem)
        {
            struct list_head *busylistp = &g_allocations.busylist;

            entry = kzalloc(sizeof(struct pool_buffer), GFP_KERNEL);
            if (!entry) {
                printk( KERN_ERR "alloc_pool_buffer failed to malloc pool_buffer struct");
                return 0;
            }

            dev_dbg( g_scmem_dev, "alloc_pool_buffer: Fits req %#llx < avail: %#llx\n", size, g_available_mem);
            entry->id = g_allocations.numbufs++;
            entry->physp = g_aligned_physstart + g_allocations.reqsize;
            // Decrease the available amount of memory
            g_available_mem -= size;
            // increase the requested amount of memory
            g_allocations.reqsize += size;
            entry->size = size;

            INIT_LIST_HEAD(&entry->dma_users);
            INIT_LIST_HEAD(&entry->uspace_users);

            dev_dbg( g_scmem_dev, "Allocated buffer %d, physical %#llx and size %#llx\n", entry->id, (unsigned long long)entry->physp, size);

            list_add_tail(&entry->element, busylistp);

            return entry;
        }

        printk( KERN_ERR "Failed to find a big enough free block\n");

        return 0;
    }

    static void releaseBuffer( struct pool_buffer* entry, struct list_head* e )
    {
        BUG_ON( e == NULL );

        if( !is_last_buffer(entry) )
        {
            list_del_init(e);
            list_add(e, &g_allocations.freelist);
            dev_dbg( g_scmem_dev, "Added buffer %#llx to freebuffer list\n", (unsigned long long) entry->physp);
        }
        else
        {
            g_available_mem += entry->size;
            g_allocations.reqsize -= entry->size;
            list_del(e);
            // Release all free buffers that haven't been released completly because of this buffer
            release_free_buffers();
            // Make sure we don't reference the freed buffer anymore if a pending dma_buf mapping is alive in some way
#ifdef CONFIG_DMA_SHARED_BUFFER
            if( entry->dmabuf )
                entry->dmabuf->priv = 0;
#endif
            dev_dbg( g_scmem_dev, "Released buffer %#llx \n", (unsigned long long) entry->physp);
            kfree(entry);
        }
    }

    static void releaseEntry( struct pool_buffer* entry, struct list_head* e )
    {
        int releaseEntry = 1;
        /* record values in case entry gets kfree()'d for CMEM_HEAP */
        int id = entry->id;

        // Check for remaining DMA mapping
        struct list_head* registeredlistp = &entry->dma_users;

        // DMA user list empty?
        if( registeredlistp->next != registeredlistp )
        {
            dev_dbg( g_scmem_dev, "Not releasing buffer %#llx (id=%d) because of pending DMA mapping \n", (unsigned long long) entry->physp, id);
            releaseEntry = 0;
        }

        // Check for remaining user allocations
        registeredlistp = &entry->uspace_users;
        if( registeredlistp->next != registeredlistp )
        {
            dev_dbg( g_scmem_dev, "Not releasing buffer %#llx (id=%d) because of pending user mapping \n", (unsigned long long) entry->physp, id);
            releaseEntry = 0;
        }

        /* no more registered users, free buffer */
        if( releaseEntry )
            releaseBuffer( entry, e );
        else
            dev_dbg( g_scmem_dev, "Buffer %#llx (id=%d) has other users left\n", (unsigned long long) entry->physp, id);
    }

    static int remove_user( struct list_head* list, void* user )
    {
        struct registered_user *reg_user;
        struct list_head *u, *unext;
        list_for_each_safe(u, unext, list)
        {
            reg_user = list_entry(u, struct registered_user, element);
            if (reg_user->user == user ) {
                dev_dbg( g_scmem_dev, "Removing user 0x%p from user list\n", reg_user->user);
                list_del(u);
                kfree(reg_user);
                break;
            }
        }
        // If no more user, return true
        return list->next == list ? 1 : 0;
    }

    // DMA_BUF implementation
#ifdef CONFIG_DMA_SHARED_BUFFER
    static int dmabuf_mmap(struct dma_buf *dmabuf, struct vm_area_struct *vma)
    {
        //struct pool_buffer *buffer = dmabuf->priv;

        int ret = 0;
        dev_dbg(g_scmem_dev, "dmabuf_mmap: vma->vm_start     = %#lx\n", vma->vm_start);
        dev_dbg(g_scmem_dev, "dmabuf_mmap: vma->vm_end       = %#lx\n", vma->vm_end);
        dev_dbg(g_scmem_dev, "dmabuf_mmap: size              = %#lx\n", vma->vm_end - vma->vm_start);
        dev_dbg(g_scmem_dev, "dmabuf_mmap: vma->vm_pgoff     = %#lx\n", vma->vm_pgoff);

        mutex_lock(&g_scmem_mutex);
        /* now map it to userspace */
        ret = mmap(vma->vm_file, vma);
        mutex_unlock(&g_scmem_mutex);

        if (ret)
            printk( KERN_ERR "%s: failure mapping buffer to userspace\n", __func__);

        return ret;
    }

//    static int dmabuf_attach(struct dma_buf *buf, struct device *dev, struct dma_buf_attachment *attachment)
//    {
//        return 0;
//    }

//    static void dmabuf_detach(struct dma_buf *buf, struct dma_buf_attachment *attachment)
//    {
//    }


    static void dmabuf_dma_buf_release(struct dma_buf *dmabuf)
    {
        struct pool_buffer *buffer = dmabuf->priv;
        if( buffer )
        {
            printk (KERN_INFO "Releasing dmabuf \n" );
            buffer->dmabuf = 0;
            WARN_ON( buffer->table.sgl != 0 );
            WARN_ON( buffer->table.nents != 0);
            WARN_ON( buffer->table.orig_nents = 0 );
        }
        else
            printk( "%s called for already released entry\n", __func__ );
    }

    static void *dmabuf_dma_buf_kmap(struct dma_buf *dmabuf, unsigned long offset)
    {
        //struct pool_buffer *buffer = dmabuf->priv;

        printk (KERN_ERR "%s entered.\n", __func__);

        return 0; //buffer->vaddr + offset * PAGE_SIZE;
    }

    static void dmabuf_dma_buf_kunmap(struct dma_buf *dmabuf, unsigned long offset, void *ptr)
    {
        printk (KERN_ERR "%s entered.\n", __func__);
    }

    static int dmabuf_dma_buf_begin_cpu_access(struct dma_buf *dmabuf, size_t start, size_t len, enum dma_data_direction direction)
    {
        //struct pool_buffer *buffer = dmabuf->priv;
        void *vaddr = 0;

        printk (KERN_ERR "%s entered.\n", __func__);

        mutex_lock(&g_scmem_mutex);
        // TODO get kernel space address (?)
        mutex_unlock(&g_scmem_mutex);
        return PTR_ERR_OR_ZERO(vaddr);
    }

    static void dmabuf_dma_buf_end_cpu_access(struct dma_buf *dmabuf, size_t start, size_t len, enum dma_data_direction direction)
    {
        //struct pool_buffer *buffer = dmabuf->priv;

        printk (KERN_ERR "%s entered.\n", __func__);

        mutex_lock(&g_scmem_mutex);
        // TODO
        mutex_unlock(&g_scmem_mutex);
    }

    static struct sg_table *dmabuf_map_dma_buf(struct dma_buf_attachment *attachment, enum dma_data_direction direction)
    {
        struct registered_user* user;
        struct dma_buf *dmabuf = attachment->dmabuf;
        struct pool_buffer *buffer = dmabuf->priv;
        struct page *virtPage;
        int ret;

        BUG_ON( dmabuf->priv == 0 );

        // Allocated already?
        if( buffer->table.sgl == 0 )
        {
            ret = sg_alloc_table(&buffer->table, 1, GFP_KERNEL);
            if (unlikely(ret) )
            {
                printk(KERN_ERR "Error allocating sg_table");
                return 0;
            }
#ifdef CONFIG_DEBUG_SG
            {
                unsigned int i;
                for (i=0; i < buffer->table.nents; i++)
                    buffer->table.sgl->sg_magic = SG_MAGIC;
            }
#endif
            virtPage = virt_to_page( phys_to_virt(buffer->physp) );
            sg_set_page(buffer->table.sgl, virtPage, PAGE_ALIGN(buffer->size), 0);
        }

        mutex_lock(&g_scmem_mutex);
        // Add user of buffer to list
        user = kmalloc(sizeof(struct registered_user), GFP_KERNEL);
        if( !user )
        {
            printk( KERN_ERR "Error allocating user for dma mapping\n");
            return 0;
        }
        user->user = attachment;
        list_add(&user->element, &buffer->dma_users);
        mutex_unlock(&g_scmem_mutex);

        dma_map_sg(attachment->dev, buffer->table.sgl, buffer->table.nents, direction);
        return &buffer->table;
    }

    // Called when the
    static void dmabuf_unmap_dma_buf(struct dma_buf_attachment *attachment, struct sg_table *table, enum dma_data_direction direction)
    {
        struct dma_buf *dmabuf = attachment->dmabuf;
        struct pool_buffer *buffer = dmabuf->priv;
        dev_dbg( g_scmem_dev, "%s entered.\n", __func__);

        BUG_ON( dmabuf->priv == 0 );

        // Release SG mapping
        dma_unmap_sg(attachment->dev, table->sgl, table->nents, direction);

        mutex_lock(&g_scmem_mutex);

        // Remove user of buffer from list
        if( remove_user( &buffer->dma_users, attachment ) )
        {
            // No remaining user space calls
            struct list_head* e = 0;

            dev_dbg( g_scmem_dev, "Releasing SG Table \n" );
            sg_free_table(&buffer->table);
            // Make sure entries are invalidated
            buffer->table.sgl = 0;
            buffer->table.nents = 0;
            buffer->table.orig_nents = 0;        

            BUG_ON( find_busy_entry( buffer->physp, 0, &e, &buffer->size ) != buffer );
            releaseEntry( buffer, e);
        }

        mutex_unlock(&g_scmem_mutex);
    }
    // end of DMA_BUF implementation
#endif // CONFIG_DMA_SHARED_BUFFER


    static int open(struct inode *inode, struct file *filp)
    {
        dev_dbg( g_scmem_dev, "open: called.\n");

        atomic_inc(&g_reference_count);

        return 0;
    }

    static int release(struct inode *inode, struct file *filp)
    {        
        int last_close = 0;

        dev_dbg( g_scmem_dev, "close: called.\n");

        /* Force free all buffers owned by the 'current' process */

        if (atomic_dec_and_test(&g_reference_count)) {
            dev_dbg( g_scmem_dev, "close: all references closed, force freeing all busy buffers.\n");
            last_close = 1;
        }

        // Check for remaining users
        if (!last_close)
            return 0;

        /* Clean up any buffers on the busy list when scmem is closed */

        /* acquire the mutex in case this isn't the last close */
        if (mutex_lock_interruptible(&g_scmem_mutex))
            return -ERESTARTSYS;


        //BUG_ON( g_allocations.busylist.next != &g_allocations.busylist );
        //BUG_ON( g_allocations.freelist.next != &g_allocations.freelist );
        if( ( g_allocations.busylist.next != &g_allocations.busylist ) )
        {
            struct pool_buffer* entry;
            struct list_head* e, *q, *busylistp = &g_allocations.busylist;
            list_for_each_safe(e, q, busylistp )
            {
                entry = list_entry(e, struct pool_buffer, element);
                dev_dbg( g_scmem_dev, "Error: Buffer %#llx still in use \n", entry->physp );
                if( entry->uspace_users.next != &entry->uspace_users )
                    dev_dbg( g_scmem_dev, "User mapping pending\n" );
                else if( entry->dma_users.next != &entry->dma_users)
                    dev_dbg( g_scmem_dev, "DMA mapping pending\n" );
            }
        }
        if( ( g_allocations.freelist.next != &g_allocations.freelist ) )
        {
            struct pool_buffer* entry;
            struct list_head* e, *q, *freelistp = &g_allocations.freelist;
            list_for_each_safe(e, q, freelistp )
            {
                entry = list_entry(e, struct pool_buffer, element);
                dev_dbg( g_scmem_dev, "Error: free Buffer %#llx still in use \n", entry->physp );
                if( entry->uspace_users.next != &entry->uspace_users )
                    dev_dbg( g_scmem_dev, "User mapping pending\n" );
                else if( entry->dma_users.next != &entry->dma_users)
                    dev_dbg( g_scmem_dev, "DMA mapping pending\n" );
            }
        }

        mutex_unlock(&g_scmem_mutex);

        dev_dbg( g_scmem_dev, "close: returning\n");

        return 0;
    }

    static int mmap(struct file *filp, struct vm_area_struct *vma)
    {
        phys_addr_t physp;
        struct pool_buffer *entry;
        unsigned long size = vma->vm_end - vma->vm_start;
        unsigned long long s;

        dev_dbg( g_scmem_dev, "mmap: vma->vm_start     = %#lx\n", vma->vm_start);
        dev_dbg( g_scmem_dev, "mmap: vma->vm_end       = %#lx\n", vma->vm_end);
        dev_dbg( g_scmem_dev, "mmap: size              = %#lx\n", size);
        dev_dbg( g_scmem_dev, "mmap: vma->vm_pgoff     = %#lx\n", vma->vm_pgoff);

        // vm_pgoff is the offset argument to the /dev/cmem file descriptor when calling mmap from userspace
        // cmem.c implements the offset to be the physical address
        // although the address is passed to mmap with PAGE_SHIFT already applied, it seems like mmap kernel implementation
        // is shifting the offset parameter by PAGE_SHIFT again (at least vm_pgoff is the offset parameter from mmap user space
        // call shifted by PAGE_SHIFT to the right).
        physp = (unsigned long long)vma->vm_pgoff << PAGE_SHIFT;

        if (mutex_lock_interruptible(&g_scmem_mutex)) {
            return -ERESTARTSYS;
        }

        s = size;
        // TODO check if size is really equal (because of page alignment)
        entry = find_busy_entry(physp, NULL, NULL, &s);

        if (entry != NULL)
        {
            if (size > entry->size)
            {
                printk( KERN_ERR "mmap: requested size %#llx too big (should be <= %#llx)\n", (unsigned long long)size, (unsigned long long)entry->size);
                mutex_unlock(&g_scmem_mutex);
                return -EINVAL;
            }

            if (entry->flags & SCMEM_CACHED)
            {
#ifdef CONFIG_ARM
                vma->vm_page_prot = __pgprot(pgprot_val(vma->vm_page_prot) |
                                             (L_PTE_MT_WRITETHROUGH | L_PTE_MT_BUFFERABLE));
#else
                // TODO find out what we have to set in case of other architectures
#endif
            }
            else
                vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

            vma->vm_flags |= VM_RESERVED | VM_IO;
            if (remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff, size, vma->vm_page_prot))
            {
                printk( KERN_ERR "mmap: failed remap_pfn_range\n");
                mutex_unlock(&g_scmem_mutex);
                return -EAGAIN;
            }

            mutex_unlock(&g_scmem_mutex);

            return 0;
        }

        printk( KERN_ERR "mmap: can't find allocated buffer with physp %#llx\n", (unsigned long long)physp);
        mutex_unlock(&g_scmem_mutex);
        return -EINVAL;

    }

    static long ioctl(struct file *filp, unsigned int cmd, unsigned long args)
    {
        unsigned long __user *argp = (unsigned long __user *) args;
        unsigned long long __user *llargp = (unsigned long long __user *) args;
        unsigned long long virtArg;
        unsigned long long physArg;
        struct list_head *e = NULL;
        struct pool_buffer *entry;
        struct registered_user *user;
        phys_addr_t physp;
        unsigned long long size = 0;
        int pool = -1;

        union SCMEM_AllocUnion allocDesc;
#ifdef CONFIG_DMA_SHARED_BUFFER
        long long dmabuf_fd = -1;
#endif

        if (_IOC_TYPE(cmd) != SCMEM_IOCMAGIC) {
            printk( KERN_ERR "ioctl(%#x): bad command type %#x (should be %#x)\n", cmd, _IOC_TYPE(cmd), SCMEM_IOCMAGIC);
            return -EINVAL;
        }

        switch (cmd)
        {
        /*
         * argp contains a pointer to an alloc descriptor coming in, and the
         * physical address and size of the allocated buffer when returning.
         */
        case SCMEM_IOC_ALLOC:
            if (copy_from_user(&allocDesc, argp, sizeof(allocDesc))) {
                return -EFAULT;
            }

            //dev_dbg( g_scmem_dev, "ALLOC%s ioctl received for memory block %zu\n", allocDesc.alloc_params.flags & SCMEM_CACHED ? "CACHED" : "", allocDesc.alloc_params.size);

            if (mutex_lock_interruptible(&g_scmem_mutex)) {
                return -ERESTARTSYS;
            }

            entry = allocBuffer( allocDesc.alloc_params.size );
            if( !entry )
            {
                mutex_unlock(&g_scmem_mutex);
                return -ENOMEM;
            }
            entry->flags = (int) allocDesc.alloc_params.flags;

            allocDesc.pool_params.physp = entry->physp;
            allocDesc.pool_params.size = entry->size;

            if (copy_to_user(argp, &allocDesc, sizeof(allocDesc))) {
                mutex_unlock(&g_scmem_mutex);
                return -EFAULT;
            }

            user = kmalloc(sizeof(struct registered_user), GFP_KERNEL);
            user->user = filp;
            list_add(&user->element, &entry->uspace_users);

            mutex_unlock(&g_scmem_mutex);

            dev_dbg( g_scmem_dev, "ALLOC%s: allocated a buffer at %#llx (phys address)\n", allocDesc.alloc_params.flags & SCMEM_CACHED ? "CACHED" : "", (unsigned long long)entry->physp);

            break;

        /*
         * argp contains either the user virtual address or the physical
         * address of the buffer to free coming in, and contains the pool
         * where it was freed from and the size of the block on return.
         */
        case SCMEM_IOC_FREE:
            //dev_dbg( g_scmem_dev, "Freeing buffer\n");

            if (copy_from_user(&allocDesc, argp, sizeof(allocDesc))) {
                return -EFAULT;
            }

            if (mutex_lock_interruptible(&g_scmem_mutex)) {
                return -ERESTARTSYS;
            }

            physp = allocDesc.pool_params.physp;
            size = allocDesc.pool_params.size;

            entry = find_busy_entry(physp, &pool, &e, &size);
            if (entry)
            {
                if( remove_user( &entry->uspace_users, filp ) )
                    releaseEntry(entry, e);
            }

            mutex_unlock(&g_scmem_mutex);

            if (!entry)
            {
                printk( KERN_ERR "Failed to free memory at %#llx\n", (unsigned long long)physp);
                return -EFAULT;
            }

            break;

        /*
         * argp contains the user virtual address of the buffer to translate
         * coming in, and the translated physical address on return.
         */
        case SCMEM_IOC_GETPHYS:

            if (get_user(virtArg, llargp))
            {
                printk( KERN_ERR "Error getting user argument\n");
                return -EFAULT;
            }

            physArg = get_phys((void *)virtArg);

            if (physArg == ~(0L)) {
                printk( KERN_ERR "GETPHYS: Failed to convert virtual %#llx to physical.\n", virtArg);
                return -EFAULT;
            }

            if (put_user(physArg, llargp)) {
                printk( KERN_ERR "Error putting user argument\n");
                return -EFAULT;
            }

            dev_dbg( g_scmem_dev, "GETPHYS: returning %#llx\n", (unsigned long long)physArg);
            break;


#ifdef CONFIG_DMA_SHARED_BUFFER
        case SCMEM_IOC_GETDMAFD:

            if( get_user(physArg, llargp) )
            {
                printk( KERN_ERR "error retrieving physical address from user space\n");
                return -EFAULT;
            }
            dev_dbg( g_scmem_dev, "SCMEM_IOC_GETDMAFD ioctl received for address: %#llx\n", physArg );

            physp = physArg;
            entry = find_busy_entry(physp, &pool, &e, 0);
            if (!entry) {
                printk (KERN_ERR "SCMEM_IOC_GETDMAFD: Cannot find pool for physical pointer %#llx.\n", physp);
                return -EFAULT;
            }

            if (!entry->dmabuf)
            {
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 17, 0)
                entry->dmabuf = dma_buf_export(entry, &dma_buf_ops, size, O_RDWR, 0);
#else
                entry->dmabuf = dma_buf_export(entry, &dma_buf_ops, size, O_RDWR);
#endif
                entry->table.sgl = 0;
                entry->table.nents= 0;
                entry->table.orig_nents = 0;
                if (IS_ERR(entry->dmabuf))
                {
                    printk (KERN_ERR "Error exporting buffer as dmabuf");
                    return -EFAULT;
                }
            }

            // TODO find out if we have to hold the FD or if we can/have to call dma_buf_fd multiple times
            if ((dmabuf_fd = dma_buf_fd(entry->dmabuf, O_CLOEXEC)) < 0) {
                printk (KERN_ERR "IOCGETFD: Failed to get dma_buf fd (err=%lld). Is it initialized?\n", dmabuf_fd);
                return (int) dmabuf_fd;
            }

            //            if (put_user(dmabuf_fd, llargp)) {
            //                printk( KERN_ERR "Error putting dma_buf to user space");
            //                return -EFAULT;
            //            }

            return (int) dmabuf_fd;
#endif
        case SCMEM_IOC_GET_MEM_PARAMS:
            allocDesc.pool_params.physp = g_aligned_physstart;
            allocDesc.pool_params.size = g_aligned_physend - g_aligned_physstart;

            if (copy_to_user(argp, &allocDesc, sizeof(allocDesc)))
                return -EFAULT;

            break;
        default:
            printk( KERN_ERR "Unknown ioctl received.\n");
            return -EINVAL;
        }

        return 0;
    }


    int __init scmem_init(void)
    {
        int err;
        unsigned long long length;
        phys_addr_t phys_end_kernel;
        struct list_head *busylistp = &g_allocations.busylist;
        struct list_head *freelistp = &g_allocations.freelist;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
        unsigned long num_physpages;
#endif

        printk(KERN_INFO "Reference Linux version %d.%d.%d\n", (LINUX_VERSION_CODE & 0x00ff0000) >> 16, (LINUX_VERSION_CODE & 0x0000ff00) >> 8, (LINUX_VERSION_CODE & 0x000000ff) >> 0 );

        mutex_init(&g_scmem_mutex);

        g_scmem_major = register_chrdev(0, "scmem", &scmem_fxns);

        if (g_scmem_major < 0) {
            printk( KERN_ERR "Failed to allocate major number.\n");
            return -ENODEV;
        }

        pr_debug("Allocated major number: %d\n", g_scmem_major);

        g_scmem_class = class_create(THIS_MODULE, "scmem");
        if (IS_ERR(g_scmem_class)) {
            printk( KERN_ERR "Error creating cmem device class.\n");
            err = -EIO;
            goto fail_after_reg;
        }

        g_scmem_dev = device_create(g_scmem_class, NULL, MKDEV(g_scmem_major, 0), NULL, "scmem");

        if (!phys_start || !phys_end)
        {
            printk( KERN_ERR "must specify both phys_start and phys_end, exiting...\n");
            err = -EINVAL;
            goto fail_after_create;
        }


        /* Get the start and end of CMEM memory */
        g_aligned_physstart = PAGE_ALIGN(simple_strtoll(phys_start, NULL, 16));
        g_aligned_physend   = PAGE_ALIGN(simple_strtoll(phys_end, NULL, 16));

        length = g_aligned_physend - g_aligned_physstart;

        if (g_aligned_physstart == 0)
        {
            printk( KERN_ERR "Physical address of 0 not allowed (phys_start%s)\n",phys_start);
            printk( KERN_ERR "  (minimum physical address is %#lx)\n", PAGE_SIZE);
            err = -EINVAL;
            goto fail_after_create;
        }

        if (g_aligned_physend < g_aligned_physstart)
        {
            printk( KERN_ERR"phys_end (%#llx) < phys_start (%#llx)\n", g_aligned_physend, g_aligned_physstart);
            err = -EINVAL;
            goto fail_after_create;
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
        num_physpages = get_num_physpages();
#endif

        /* attempt to determine the end of Linux kernel memory */
        phys_end_kernel = virt_to_phys((void *)PAGE_OFFSET) + (num_physpages << PAGE_SHIFT);

        if(phys_end_kernel > g_aligned_physstart)
        {
            if (allowOverlap == 0) {
                printk( KERN_ERR "SCMEM phys_start (%#llx) overlaps kernel (%#llx -> %#llx)\n", g_aligned_physstart, (unsigned long long)virt_to_phys((void *)PAGE_OFFSET), (unsigned long long)phys_end_kernel);
                err = -EINVAL;
                goto fail_after_create;
            }
            else {
                printk( KERN_WARNING "SCMEM Range Overlaps Kernel Physical - allowing overlap\n");
                printk( KERN_WARNING "SCMEM phys_start (%#llx) overlaps kernel (%#llx -> %#llx)\n", g_aligned_physstart, (unsigned long long)virt_to_phys((void *)PAGE_OFFSET), (unsigned long long)phys_end_kernel);
            }
        }

        /* Initialize the top memory chunk in which to put the pools */
        dev_dbg( g_scmem_dev, "Phys_end_kernel = %#llx\n", (unsigned long long) phys_end_kernel);
        dev_dbg( g_scmem_dev, "calling request_mem_region(%#llx-%llx, %lld, \"SCMEM\")\n", g_aligned_physstart, g_aligned_physend, length);

        // request_mem_region tells the kernel that our driver is going to use this range of I/O addresses,
        // which will prevent other drivers to make any overlapping call to the same region through request_mem_region.
        // This mechanism does not do any kind of mapping, it's a pure reservation mechanism, which relies on the fact
        // that all kernel device drivers must be nice, and they must call request_mem_region, check the return value,
        // and behave properly in case of error.
        if (!request_mem_region(g_aligned_physstart, length, "SCMEM"))
        {
            printk( KERN_ERR "Failed to request_mem_region(%#llx, %#llx)\n", g_aligned_physstart, length);
            err = -EFAULT;
            goto fail_after_create;
        }
        else {
            g_mem_flags |= BLOCK_MEMREGION;
        }

        g_available_mem = length;

        INIT_LIST_HEAD(busylistp);
        INIT_LIST_HEAD(freelistp);

        g_allocations.numbufs = 0;
        g_allocations.reqsize = 0;

        printk(KERN_INFO "scmemk initialized\n");

        return 0;

fail_after_create:

        if( g_mem_flags & BLOCK_MEMREGION )
        {
            length = g_aligned_physend - g_aligned_physstart;
            dev_dbg( g_scmem_dev, "calling release_mem_region(%#llx, %#llx)...\n", g_aligned_physstart, length);
            release_mem_region(g_aligned_physstart, length);

            g_mem_flags &= ~BLOCK_MEMREGION;
        }

        device_destroy(g_scmem_class, MKDEV(g_scmem_major, 0));
        class_destroy(g_scmem_class);

fail_after_reg:
        printk( KERN_WARNING "Unregistering character device scmem\n");
        unregister_chrdev(g_scmem_major, "scmem");

        return err;
    }


    void __exit scmem_exit(void)
    {        
        unsigned long long length = g_aligned_physend - g_aligned_physstart;

        dev_dbg(g_scmem_dev, "In scmem_exit()\n");

        BUG_ON( g_allocations.busylist.next != &g_allocations.busylist );
        BUG_ON( g_allocations.freelist.next != &g_allocations.freelist );

        if (g_mem_flags & BLOCK_MEMREGION) {
            dev_dbg(g_scmem_dev, "calling release_scmem_region(%#llx-%#llx)...\n", g_aligned_physstart, g_aligned_physend);
            release_mem_region(g_aligned_physstart, length);
            g_mem_flags &= ~BLOCK_MEMREGION;
        }

        device_destroy(g_scmem_class, MKDEV(g_scmem_major, 0));
        class_destroy(g_scmem_class);

        printk( KERN_DEBUG "Unregistering character device scmem\n");
        unregister_chrdev(g_scmem_major, "scmem");

        printk(KERN_INFO "scmemk unregistered\n");
    }

    MODULE_LICENSE("GPL");
    module_init(scmem_init);
    module_exit(scmem_exit);
