#include "scmem.h"
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>
#include <string.h>
#include "../driver/scmemk.h"
#include <stdio.h>

#define EMSG(fmt, args...) ((void) (fprintf( stderr, "%s(%d): " fmt "\n", __FILE__, __LINE__, ## args) && fflush(stderr)))
#define SMSG(fmt, args...) ((void) (fprintf( stdout, fmt "\n", ## args) && fflush(stdout)))

// TODO make sure this also works with 64bit address

static int g_scmem_fd = 0;
static int g_ref_count = 0;

int scmem_init(void)
{
    if (g_scmem_fd > 0)
    {
        ++g_ref_count;
        SMSG( "init: /dev/cmem already opened, incremented ref_count %d\n", g_ref_count);
        return 0;
    }

    g_scmem_fd = open("/dev/scmem", O_RDWR | O_CLOEXEC);

    if (g_scmem_fd == -1) {
        EMSG("init: Failed to open /dev/scmem: '%s'\n", strerror(errno));
        return -1;
    }

    g_ref_count++;

    return 0;
}

int scmem_deinit(void)
{
    g_ref_count--;

    if( g_ref_count == 0 )
    {
        close( g_scmem_fd );
        g_scmem_fd = 0;
    }
    else if( g_ref_count < 0 )
    {
        EMSG( "Unbalanced call to scmem_init and scmem_deinit");
    }
    return g_ref_count;
}

int scmem_get_mem_params(unsigned long long *phys_start, size_t *size)
{
    union SCMEM_AllocUnion params;
    int err;

    if( !g_scmem_fd )
    {
        EMSG("SCMEM not initialized");
        return -1;
    }

    if( !phys_start || !size )
    {
        EMSG("Invalid parameters given");
        return -1;
    }

    params.pool_params.physp = 0;
    params.pool_params.size = 0;

    if( (err = ioctl(g_scmem_fd, SCMEM_IOC_GET_MEM_PARAMS, &params)) != 0 )
        EMSG( "Error in ioctl: %d", err);

    *phys_start = params.pool_params.physp;
    *size = params.pool_params.size;
    return err;
}

void *scmem_map(unsigned long long physp, size_t size)
{
    void *userp;

    if( !g_scmem_fd )
    {
        EMSG("SCMEM not initialized");
        return NULL;
    }

    SMSG("map: mapping phys buffer %#llx, size %ld\n", (unsigned long long)physp, size);

    /* Map the physical address to user space */
    userp = mmap(0,                       // Preferred start address
                 size,                    // Length to be mapped
                 PROT_WRITE | PROT_READ,  // Read and write access
                 MAP_SHARED,              // Shared memory
                 g_scmem_fd,                 // File descriptor
                 physp);               // The byte offset from fd

    if (userp == MAP_FAILED) {
        EMSG("map: Failed to mmap buffer at physical address %#llx\n",
            (unsigned long long)physp);

        return NULL;
    }

    SMSG("map: mmap succeeded, returning virt buffer 0x%p\n", userp);

    return userp;
}

void scmem_unmap(void* userPtr, size_t size)
{
    if( !g_scmem_fd )
    {
        EMSG("SCMEM not initialized");
        return;
    }

    if( munmap(userPtr, size) != 0 )
        EMSG("Error %d, during unmap", errno );
}

int scmem_dma_fd(unsigned long long phys)
{
    int err = 0;

    if( !g_scmem_fd )
    {
        EMSG("SCMEM not initialized");
        return 0;
    }

    if( (err = ioctl(g_scmem_fd, SCMEM_IOC_GETDMAFD, &phys)) <= 0 )
    {
        EMSG( "Error %d retrieving dma_buf file descriptor!", err);
        return 0;
    }

    return err;
}

off_t scmem_alloc(size_t size, int cached)
{
    union SCMEM_AllocUnion params;
    int err;
    if( !g_scmem_fd )
    {
        EMSG("SCMEM not initialized");
        return 0;
    }

    params.alloc_params.size = size;
    params.alloc_params.flags = cached ? SCMEM_CACHED : 0;

    if( (err = ioctl(g_scmem_fd, SCMEM_IOC_ALLOC, &params)) == 0 )
    {
        return params.pool_params.physp;
    }

    EMSG( "Error in ioctl: %d", err);
    return 0;
}


void scmem_free(unsigned long long phys )
{
    union SCMEM_AllocUnion params;
    int err;
    if( !g_scmem_fd )
    {
        EMSG("SCMEM not initialized");
        return;
    }

    params.pool_params.physp = phys;
    params.pool_params.size = 0;

    if( (err = ioctl(g_scmem_fd, SCMEM_IOC_FREE, &params)) != 0 )
        EMSG( "Error in ioctl: %d", err);

}
