#ifndef SCMEM_H_
#define SCMEM_H_

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief retrieve informations about the memory managed by scmem
 * @param phys_start (out) filled by this function with the physical start address of the memory managed by scmem
 * @param size (out) filled by this function with the whole size of the memory managed by scmem
 */
extern int scmem_get_mem_params(unsigned long long *phys_start, size_t* size );

/**
 * @brief map allocated physical block into user space
 * @param physp address of buffer allocated via scmem_alloc()
 * @param size size of memory to be mapped
 * @return user space address or 0 in case of an error
 */
extern void *scmem_map(unsigned long long physp, size_t size);

/**
 * @brief unmap userspace mapping
 * @param userPtr pointer retrieved via scmem_map
 * @param size size requested when scmem_map() was called
 */
extern void scmem_unmap(void* userPtr, size_t size);

/**
 * @brief get dmabuf API file descriptor for buffer
 * @param phys address of buffer allocated via scmem_alloc
 * @return negative error code on failure or a file descriptor number on success
 */
extern int scmem_dma_fd(unsigned long long phys);

/**
 * @brief open /dev/scmem device
 * can be called multiple times without side-effects, but scmem_deinit() has to be called as often as init has been called
 * @return 0 on success, -1 in case of an error
 */
extern int scmem_init(void);

/**
 * @brief closes /dev/scmem file descriptor
 * has to be called as often as scmem_init() has been called before
 * @return reference count number (scmem_init calls() with no matching scmem_deinit() call) left
 */
extern int scmem_deinit(void);

/**
 * @brief allocates a block of memory and returns the physical address to this block
 * @param size number of bytes to allocate
 * @param cached if set, during mapping into user space the mapping will be cachable, if 0, pgprot_noncached will be used
 * @return 0 in case of an error, physical address to memory in case of success
 */
extern off_t scmem_alloc(size_t size, int cached);

/**
 * @brief frees a block of memory allocated before by using scmem_alloc()
 * @param phys the phsyical address to free
 * @param size size of the memory (has to be equal to the one specified when calling scmem_alloc)
 */
extern void scmem_free(unsigned long long phys);

#ifdef __cplusplus
}
#endif

#endif
